@@ -3,27 +3,31 @@ from __future__ import annotations
 import random
 from datetime import date, datetime, time, timedelta
 from typing import List, Union, Optional, Dict, Tuple
+from warnings import warn
 
 from .elements_base import _Action, Value, MultiChoiceInput
 from .elements_base import Element, MediaElement, InputElement, \
-                           ActionChoiceInput, ChoiceInput, ChoiceInput1D, \
-                           OtherChoiceInput, TextInput, \
-                           DateElement, Grid, TimeElement
+                           ActionChoiceInput, ChoiceInput1D, SingleInput, \
+                           OtherChoiceInput, ValidatedInput, TextInput, \
+                           DateInput, Grid, TimeInput
 from .elements_base import ChoiceValue, EmptyValue, MultiChoiceValue, TextValue, \
                            GridChoiceValue, GridMultiChoiceValue
 
 from .errors import ElementTypeError, InvalidDuration, \
-    RequiredElement, InvalidText, MisconfiguredGrid, UnknownValidator
+                    InvalidText, MisconfiguredElement, \
+                    UnknownValidator, InvalidArguments, UnknownElement
 from .options import ActionOption
-from .util import RADIO_SYMBOLS, CHECKBOX_SYMBOLS, add_indent, elem_separator, random_subset
-from .validators import GridValidator
+from .util import RADIO_SYMBOLS, CHECKBOX_SYMBOLS, add_indent, elem_separator, random_subset, \
+    list_get
+from .validators import CheckboxValidator, CheckboxTypes
 
 
-__all__ = ['Element', '_Action', 'CallbackRetVal', 'Value',
-           'Page', 'Comment', 'Image', 'Video',
+__all__ = ['CallbackRetVal', 'Value',
+           'Unknown', 'UserEmail', 'Page', 'Comment', 'Image', 'Video',
            'Short', 'Paragraph', 'Radio', 'Dropdown', 'Checkboxes', 'Scale',
            'RadioGrid', 'CheckboxGrid',
            'Time', 'Duration', 'Date', 'DateTime',
+           'FileUpload',
            'default_callback', 'parse']
 
 
@@ -78,10 +82,8 @@ class Page(Element):
     @classmethod
     def _parse(cls, elem):
         res = super()._parse(elem)
-        res['prev_action'] = _Action.NEXT
-        if len(elem) > cls._Index.ACTION:
-            # FIRST / NEXT / SUBMIT / page id
-            res['prev_action'] = elem[cls._Index.ACTION]
+        # FIRST / NEXT / SUBMIT / page id
+        res['prev_action'] = list_get(elem, cls._Index.ACTION, _Action.NEXT)
         return res
 
     def __init__(self, *, prev_action, **kwargs):
@@ -101,7 +103,6 @@ class Page(Element):
 
     def next_page(self):
         """The next pge, based on user's choices."""
-
         if self._next_page is None:
             return None
 
@@ -113,7 +114,6 @@ class Page(Element):
 
     def to_str(self, indent=0, include_answer=False):
         """See base class."""
-
         title = f'Page {self.index + 1}:'
         if self.name:
             title = f'{title} {self.name}'
@@ -138,7 +138,6 @@ class Page(Element):
 
     def payload(self) -> Dict[str, List[str]]:
         """Returns a combined payload of all elements from this page."""
-
         payload = {}
         for elem in self.elements:
             if isinstance(elem, InputElement):
@@ -147,14 +146,13 @@ class Page(Element):
 
     def draft(self) -> List[Tuple]:
         """Returns a combined draftResponse."""
-
         result = []
         for elem in self.elements:
             if isinstance(elem, InputElement):
                 result += elem.draft()
         return result
 
-    def _resolve_actions(self, next_page: Page, mapping):
+    def _resolve_actions(self, next_page: Optional[Page], mapping):
         for elem in self.elements:
             if isinstance(elem, ActionChoiceInput):
                 elem._resolve_actions(next_page, mapping)
@@ -176,6 +174,10 @@ Page.SUBMIT = Page(id_=_Action.SUBMIT, name=None, description=None,
                    type_=Element.Type.PAGE, prev_action=None).with_index(_Action.SUBMIT)
 
 
+class Unknown(Element):
+    pass
+
+
 class Comment(Element):
     pass
 
@@ -221,7 +223,41 @@ class Short(TextInput):
             return
         value = self._value[0]
         if '\n' in value:
-            raise InvalidText(self, value.replace('\n', r'\n'), 'Input contains newlines')
+            raise InvalidText(self, value.replace('\n', r'\n'), details='Input contains newlines')
+
+
+class UserEmail(Short):
+    """A virtual element used for e-mail collection."""
+
+    ID = -9999
+    ENTRY_ID = -999
+
+    def __init__(self):
+        from .validators import TextValidator, TextTypes
+        val = TextValidator(type_=TextValidator.Type.TEXT, subtype=TextTypes.EMAIL,
+                            error_msg='', args=None, bad_args=False)
+        super().__init__(
+            id_=self.ID,
+            name='Email',
+            description='Your email',
+            type_=Element.Type.SHORT,
+            entry_ids=[self.ENTRY_ID],
+            required=True,
+            validator=val
+        )
+
+    def _hints(self, indent=0, modify=False):
+        # Don't include the validator hint
+        from .elements_base import ValidatedInput
+        return super(ValidatedInput, self)._hints(indent, modify)
+
+    def payload(self):
+        """See base class."""
+        return {'emailAddress': self._value}
+
+    def draft(self):
+        """Returns empty list for Page.draft."""
+        return []
 
 
 class Paragraph(TextInput):
@@ -243,9 +279,19 @@ class Dropdown(ActionChoiceInput):
     pass
 
 
-class Checkboxes(OtherChoiceInput, MultiChoiceInput):
+class Checkboxes(ValidatedInput, OtherChoiceInput, MultiChoiceInput):
     _choice_symbols = CHECKBOX_SYMBOLS
 
+    class _Index(ChoiceInput1D._Index):
+        VALIDATOR = 4
+
+    @classmethod
+    def _parse_validator(cls, elem) -> Optional[CheckboxValidator]:
+        value = cls._get_entry(elem)
+        if len(value) > cls._Index.VALIDATOR and value[cls._Index.VALIDATOR]:
+            return CheckboxValidator.parse(value[cls._Index.VALIDATOR][0])
+        return None
+
     def set_value(self, value: Union[CheckboxesValue, EmptyValue]):
         """Sets the value for this element.
 
@@ -259,9 +305,18 @@ class Checkboxes(OtherChoiceInput, MultiChoiceInput):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         return self._set_choices([self._to_choice_list(value)])
 
+    def _is_misconfigured(self):
+        cnt = len(self.options)
+        if self.other_option is not None:
+            cnt += 1
+        required_cnt = self.validator.args[0]  # always > 0
+        if (self.validator.subtype is CheckboxTypes.AT_LEAST
+                or self.validator.subtype is CheckboxTypes.EXACTLY) and cnt < required_cnt:
+            return True
+        return False
+
 
 class Scale(ChoiceInput1D):
     """A liner scale.
@@ -271,7 +326,7 @@ class Scale(ChoiceInput1D):
         high: The label for the highest value (may be empty).
     """
 
-    class _Index(ChoiceInput._Index):
+    class _Index(ChoiceInput1D._Index):
         LABELS = 3
 
     @classmethod
@@ -301,7 +356,6 @@ class Scale(ChoiceInput1D):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         if isinstance(value, int):
             value = str(value)
         return self._set_choices([self._to_choice_list(value)])
@@ -343,10 +397,6 @@ class RadioGrid(Grid):
                 The argument is a list,
                 but its length does not match the number of rows.
         """
-
-        # NOTE Maybe it's better to allow list of lists with lengths <= 1
-        # grid.set_value([[val1], [], [val3], ...])
-        # instead of grid.set_value([val1, Value.EMPTY, val3, ...])
         self._set_grid_values(value)
 
 
@@ -369,11 +419,10 @@ class CheckboxGrid(Grid, MultiChoiceInput):
                 The argument is a list,
                 but its length does not match the number of rows.
         """
-
         self._set_grid_values(value)
 
 
-class Time(TimeElement):
+class Time(TimeInput):
     def set_value(self, value: Union[TimeValue, EmptyValue]):
         """Sets the value for this element.
 
@@ -386,7 +435,6 @@ class Time(TimeElement):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         if value is Value.EMPTY:
             return self._set_time(None, None, None)
         if isinstance(value, time):
@@ -399,7 +447,7 @@ class Time(TimeElement):
         return [f'{self._hour:02}:{self._minute:02}']
 
 
-class Duration(TimeElement):
+class Duration(TimeInput):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self._timedelta = None
@@ -416,7 +464,6 @@ class Duration(TimeElement):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         if value is Value.EMPTY:
             self._set_time(None, None, None)
             self._timedelta = None
@@ -428,10 +475,8 @@ class Duration(TimeElement):
         else:
             raise ElementTypeError(self, value)
 
-    def validate(self):
-        """See base class."""
-
-        super().validate()
+    def _validate(self):
+        super()._validate()
         if self._timedelta is None:
             return
         seconds = self._timedelta.total_seconds()
@@ -447,7 +492,7 @@ class Duration(TimeElement):
         return [f'{h:02}:{m:02}:{s:02}']
 
 
-class Date(DateElement):
+class Date(DateInput):
     def set_value(self, value: Union[DateValue, EmptyValue]):
         """Sets the value for this element.
 
@@ -460,7 +505,6 @@ class Date(DateElement):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         if value is Value.EMPTY:
             self._date = None
         elif isinstance(value, date):
@@ -469,7 +513,7 @@ class Date(DateElement):
             raise ElementTypeError(self, value)
 
 
-class DateTime(DateElement):
+class DateTime(DateInput):
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
         self._time = None
@@ -486,19 +530,18 @@ class DateTime(DateElement):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         if value is Value.EMPTY:
             date_ = time_ = None
         elif isinstance(value, datetime):
             date_, time_ = value.date(), value.time()
         else:
             raise ElementTypeError(self, value)
+        # self._date is None if and only if self._time is None, no need to extend validation
         self._date = date_
         self._time = time_
 
     def payload(self) -> Dict[str, List[str]]:
         """See base class."""
-
         if self._date is None:
             return {}
         payload = super().payload()
@@ -510,7 +553,6 @@ class DateTime(DateElement):
 
     def draft(self) -> List[Tuple]:
         """See base class."""
-
         if self._date is None:
             return []
         result = super().draft()
@@ -518,13 +560,6 @@ class DateTime(DateElement):
         result[0][2][0] = date_str + f' {self._time.strftime("%H:%M")}'
         return result
 
-    def validate(self):
-        """See base class."""
-
-        super().validate()
-        if self.required and self._time is None:
-            raise RequiredElement(self)
-
     def _answer(self) -> List[str]:
         if self._date is None:
             return ['EMPTY']
@@ -533,53 +568,86 @@ class DateTime(DateElement):
         return answer
 
 
-def _val_grid_choices(elem):
-    if elem.is_misconfigured():
-        raise MisconfiguredGrid(elem)
-    if elem.validator.type == GridValidator.Type.UNKNOWN:
-        # Raise, not warn. Otherwise, most probably, form submission will fail
-        # If needed, custom callback should handle such cases
-        raise UnknownValidator(type(elem.validator), elem.validator.type)
-    n = len(elem.rows)
+class FileUpload(SingleInput):
+    # As of June 2021, forms with a file upload require sign in for loading,
+    # so this class is actually never used
+    def set_value(self, value):
+        raise NotImplementedError()
 
+
+def _grid_validated_choices(elem: Grid):
+    n = len(elem.rows)
     if isinstance(elem, RadioGrid):
-        if len(elem.options) >= n:
+        if elem.required:
+            # cols >= rows (checked in default_callback by elem.is_misconfigured)
+            # Just pick a unique option for each row.
             return random.sample(elem.options, k=n)
-        # cols < rows, not required
-        row_indices = random.sample(range(n), k=len(elem.options))
+        # Pick some options and put each into a unique random row.
+        # With prob=0.5 the grid is too empty.
+        opts = random_subset(elem.options, max_size=n, prob=0.8)
+        row_indices = random.sample(range(n), k=len(opts))
         choices = [Value.EMPTY for _ in range(n)]
-        for opt, idx in zip(elem.options, row_indices):
+        for opt, idx in zip(opts, row_indices):
             choices[idx] = opt
         return choices
 
-    if not elem.required:
-        row_indices = random.choices(range(n), k=len(elem.options))
-        choices = [[] for _ in range(n)]
-        for opt, idx in zip(elem.options, row_indices):
-            choices[idx].append(opt)
-        return choices
-    # Required CheckboxGrid with unique columns
-    choices = random.sample(elem.options, k=n)
-    res = [[choice] for choice in choices]
-    remaining = set(elem.options) - set(choices)
-    if not remaining:
-        return res
-    row_indices = random.choices(range(n), k=len(remaining))
-    for opt, idx in zip(remaining, row_indices):
-        res[idx].append(opt)
-    return res
+    # CheckboxGrid
+    result = [[] for _ in range(n)]
+    opts = random_subset(elem.options,
+                         min_size=n if elem.required else 0,
+                         prob=0.8)
+    if elem.required:
+        result = [[opt] for opt in opts[:n]]  # choose at least one option in each row
+        opts = opts[n:]
+
+    # Add each of the remaining options to a random row
+    row_indices = random.choices(range(n), k=len(opts))
+    for opt, idx in zip(opts, row_indices):
+        result[idx].append(opt)
+    return result
+
+
+def _cb_validated_choices(elem: Checkboxes):
+    required_cnt = elem.validator.args[0]
+    if elem.validator.subtype is CheckboxTypes.AT_MOST:
+        return random_subset(elem.options, min_size=int(elem.required), max_size=required_cnt)
+    if len(elem.options) + int(elem.other_option is not None) < required_cnt:
+        # Impossible to select the required number of options.
+        # Element is optional, this is the only valid value
+        return Value.EMPTY
+    if len(elem.options) < required_cnt:
+        if not elem.required:
+            return Value.EMPTY
+        raise NotImplementedError('Need to choose "Other" value to meet validation requirements')
+    if elem.validator.subtype is CheckboxTypes.AT_LEAST:
+        return random_subset(elem.options, min_size=required_cnt)
+    if elem.validator.subtype is CheckboxTypes.EXACTLY:
+        return random.sample(elem.options, required_cnt)
+
+
+def _validated_choices(elem):
+    if isinstance(elem, Grid):
+        return _grid_validated_choices(elem)
+    if isinstance(elem, Checkboxes):
+        return _cb_validated_choices(elem)
+    raise NotImplementedError()
 
 
 def default_callback(elem: InputElement, page_index, elem_index) -> Union[ElemValue, EmptyValue]:
-    """The default callback implementation for Form.fill."""
+    """The default callback implementation for Form.fill.
 
-    if isinstance(elem, Scale) or isinstance(elem, Dropdown):
-        return random.choice(elem.options)
-    if isinstance(elem, Radio):
-        # Don't auto-choose "Other"
-        return random.choice(elem.options)
+    This callback will raise a NotImplementedError
+    if it is called on a TextInput, DateInput or a TimeInput.call
+    """
 
-    # Example: allow "Other"
+    # Single choice inputs
+    if isinstance(elem, Scale) or isinstance(elem, Dropdown) or isinstance(elem, Radio):
+        opts = elem.options[:]
+        if not elem.required:
+            opts.append(Value.EMPTY)
+        return random.choice(opts)
+
+    # Example: allow "Other" in Radio
     #     if elem.other_option is not None:
     #         opt = random.choice(elem.options + [elem.other_option])
     #         if opt.other:
@@ -587,18 +655,33 @@ def default_callback(elem: InputElement, page_index, elem_index) -> Union[ElemVa
     #             opt.value = 'Sample text'
     #         return opt
 
+    # Check the validator; for TextInput raise NotImplementedError (later)
+    if isinstance(elem, ValidatedInput) and not isinstance(elem, TextInput) \
+            and elem.validator is not None:
+
+        if elem.is_misconfigured():
+            raise MisconfiguredElement(elem)
+        if elem.validator.has_unknown_type():
+            # Raise, not warn. Otherwise, most probably, form submission will fail
+            # If needed, custom callback should handle such cases
+            raise UnknownValidator(
+                type(elem.validator), [elem.validator.type, elem.validator.subtype]
+            )
+        if elem.validator.bad_args:
+            raise InvalidArguments(
+                type(elem.validator), elem.validator.type,
+                elem.validator.subtype, elem.validator.args
+            )
+        return _validated_choices(elem)
+
     if isinstance(elem, Checkboxes):
         # Don't auto-choose "Other"
-        return random_subset(elem.options, nonempty=elem.required)
+        return random_subset(elem.options, min_size=int(elem.required))
 
     if isinstance(elem, Grid):
-        if elem.validator is not None:
-            # NOTE this function will choose maximal allowed number of columns,
-            # even if required is False (find a better solution or disable this feature entirely?)
-            return _val_grid_choices(elem)
         n = len(elem.rows)
         if isinstance(elem, CheckboxGrid):
-            return [random_subset(elem.options, nonempty=elem.required) for _ in range(n)]
+            return [random_subset(elem.options, min_size=int(elem.required)) for _ in range(n)]
         opts = elem.options
         if not elem.required:
             opts = opts + [Value.EMPTY]
@@ -609,22 +692,26 @@ def default_callback(elem: InputElement, page_index, elem_index) -> Union[ElemVa
 
 
 _element_mapping = {getattr(Element.Type, el_type.__name__.upper()): el_type for el_type in [
+    Unknown,
     Short, Paragraph,
     Radio, Dropdown, Checkboxes, Scale,
     Comment, Page, Image, Video,
 ]}
 
+_element_mapping[Element.Type.FILE_UPLOAD] = FileUpload
+
 
 def parse(elem):
     """Creates an Element of the right type from its JSON representation."""
-
     el_type = Element.Type(elem[Element._Index.TYPE])
     if el_type is Element.Type.GRID:
         cls = CheckboxGrid if Grid.parse_multichoice(elem) else RadioGrid
     elif el_type is Element.Type.DATE:
-        cls = DateTime if DateElement.parse_time_flag(elem) else Date
+        cls = DateTime if DateInput.parse_time_flag(elem) else Date
     elif el_type is Element.Type.TIME:
-        cls = Duration if TimeElement.parse_duration_flag(elem) else Time
+        cls = Duration if TimeInput.parse_duration_flag(elem) else Time
     else:
         cls = _element_mapping[el_type]
+    if el_type is Element.Type.UNKNOWN:
+        warn(UnknownElement(elem, elem[Element._Index.TYPE]))
     return cls.parse(elem)
