@@ -3,24 +3,30 @@ from __future__ import annotations
 from abc import ABC, abstractmethod
 
 from enum import Enum, auto
-from typing import Dict, List, Literal, Union, cast, Any, Optional, Tuple
+from typing import Dict, List, Literal, Union, cast, Any, Optional, Tuple, TYPE_CHECKING
 
-from .validators import GridValidator, TextValidator
+from .util import DefaultEnum, list_get
+from .validators import Validator, GridValidator, TextValidator, GridTypes, NumberTypes
 from .errors import DuplicateOther, EmptyOther, InvalidChoice, \
     ElementTypeError, ElementValueError, RequiredElement, RequiredRow, RowTypeError, \
-    InvalidRowChoice
+    InvalidRowChoice, MisconfiguredElement
 from .options import ActionOption, Option, parse as parse_option
 
+if TYPE_CHECKING:
+    from .form import Form
+
 
 class Value(Enum):
     """A value which can be returned by a callback in Form.fill
 
     DEFAULT: Use the value returned by the default callback.
     EMPTY: Leave the element empty.
+    UNCHANGED: Leave the element value unchanged.
     """
 
     DEFAULT = auto()
     EMPTY = auto()
+    UNCHANGED = auto()
 
 
 EmptyValue = Literal[Value.EMPTY]
@@ -55,8 +61,8 @@ class Element:
         type: The element type.
     """
 
-    class Type(Enum):
-        # NOTE File upload element is not implemented
+    class Type(DefaultEnum):
+        UNKNOWN = -1
         SHORT = 0
         PARAGRAPH = 1
         RADIO = 2
@@ -70,6 +76,7 @@ class Element:
         TIME = 10
         IMAGE = 11
         VIDEO = 12
+        FILE_UPLOAD = 13
 
     class _Index:
         ID = 0
@@ -84,7 +91,6 @@ class Element:
         This method should not be called directly,
         use gforms.elements.parse instead.
         """
-
         return cls(**cls._parse(elem))
 
     @classmethod
@@ -101,13 +107,16 @@ class Element:
         self.name = name
         self.description = description
         self.type = type_
+        self._form: Optional['Form'] = None
+
+    def bind(self, form):
+        self._form = form
 
     def to_str(self, indent=0, **kwargs):
         """Returns a text representation of the element.
 
         For args description, see Form.to_str.
         """
-
         s = f'{self._type_str()}'
         if self.name:
             s = f'{s}: {self.name}'
@@ -120,12 +129,15 @@ class Element:
 
 
 class InputElement(Element, ABC):
-    """An element which can be filled."""
+    """An element which can be filled.
+
+    Attributes:
+        required: Self-explanatory.
+    """
 
     class _Index(Element._Index):
         ENTRIES = 4
         ENTRY_ID = 0
-        OPTIONS = 1
         REQUIRED = 2
 
     @classmethod
@@ -134,7 +146,7 @@ class InputElement(Element, ABC):
         entries = cls._get_entries(elem)
         res.update({
             'entry_ids': [e[cls._Index.ENTRY_ID] for e in entries],
-            'required': entries[0][cls._Index.REQUIRED],  # same for all entries
+            'required': bool(entries[0][cls._Index.REQUIRED]),  # same for all entries
         })
         return res
 
@@ -158,7 +170,6 @@ class InputElement(Element, ABC):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         raise NotImplementedError()
 
     def validate(self):
@@ -167,16 +178,18 @@ class InputElement(Element, ABC):
         Raises a ValidationError if the value is invalid.
 
         Raises:
-            gforms.errors.ValidationError:
-                The value set for this element is invalid.
+            gforms.errors.RequiredElement:
+                This element is required, but the value is empty.
+            gforms.errors.InvalidValue:
+                The value has a correct type,
+                but is not valid for this element.
         """
-
-        for i in range(len(self._values)):
-            self._validate_entry(i)
+        self._validate()
+        if self._form is not None:
+            self._form._unvalidated_elements.discard(self)
 
     def to_str(self, indent=0, include_answer=False):
         """See base class."""
-
         parts = self._header()
         hints = self._hints(indent, include_answer)
         if hints:
@@ -195,7 +208,6 @@ class InputElement(Element, ABC):
         The payload contains the element's value(s)
         and can be used as a (part of) POST request body.
         """
-
         payload = {}
         for entry_id, value in zip(self._entry_ids, self._values):
             if value:
@@ -204,7 +216,6 @@ class InputElement(Element, ABC):
 
     def draft(self) -> List[Tuple]:
         """Returns a (part of) emulated draftResponse for Form.submit."""
-
         result = []
         for entry_id, value in zip(self._entry_ids, self._values):
             if value:
@@ -216,6 +227,14 @@ class InputElement(Element, ABC):
                 ))
         return result
 
+    def prefill(self, prefilled_data: Dict[str, List[str]]):
+        """Fills the element with values from the prefilled link."""
+        values = [prefilled_data.get(entry_id, []) for entry_id in self._entry_ids]
+        # prefilled data still needs to be validated
+        # (e.g. a modified url was used or elements were updated)
+        # Also, there may be no prefilled values for a required element.
+        self._set_values(values)
+
     @staticmethod
     def _submit_id(entry_id):
         return f'entry.{entry_id}'
@@ -224,6 +243,9 @@ class InputElement(Element, ABC):
     def _get_entries(elem):
         return elem[InputElement._Index.ENTRIES]
 
+    def _is_empty(self):
+        return not any(self._values)
+
     def _set_values(self, values: List[Union[List[str], EmptyValue]]):
         values = values[:]
         for i, entry_values in enumerate(values):
@@ -231,6 +253,8 @@ class InputElement(Element, ABC):
                 values[i] = []
 
         self._values = values  # type: ignore
+        if self._form is not None:
+            self._form._unvalidated_elements.add(self)
 
     def _header(self) -> List[str]:
         parts = [self._type_str()]
@@ -247,7 +271,6 @@ class InputElement(Element, ABC):
 
         Returned strings should be already indented.
         """
-
         return []
 
     def _answer(self) -> List[str]:
@@ -259,6 +282,10 @@ class InputElement(Element, ABC):
             return 'EMPTY'
         return ', '.join(f'"{value}"' for value in entry_val)
 
+    def _validate(self):
+        for i in range(len(self._values)):
+            self._validate_entry(i)
+
     def _validate_entry(self, index):
         if self.required and not self._values[index]:
             raise RequiredElement(self, index=index)
@@ -293,6 +320,9 @@ class ChoiceInput(InputElement, ABC):
         options: A list of allowed choices for this element.
     """
 
+    class _Index(InputElement._Index):
+        OPTIONS = 1
+
     # This class is basically the same as CheckboxGrid
     _has_multichoice = False
 
@@ -330,9 +360,8 @@ class ChoiceInput(InputElement, ABC):
                 )
         self._set_values(new_choices)
 
-    @abstractmethod
     def _hints(self, indent=0, modify=False):
-        raise NotImplementedError()
+        return []
 
     def _find_option(self, value: ChoiceValue, i):
         if isinstance(value, Option):
@@ -375,11 +404,31 @@ class MultiChoiceInput(ChoiceInput):
 
 
 class ChoiceInput1D(ChoiceInput, SingleInput):
-    """A ChoiceInput with a single entry."""
+    """A ChoiceInput with a single entry.
+
+    Attributes:
+        shuffle_options: Self-explanatory..
+    """
 
     # ChoiceInput1D == Checkboxes without "other"
     _choice_symbols = ('·', '>')
 
+    class _Index(ChoiceInput._Index):
+        SHUFFLE_OPTIONS = 8  # Does not affect Scale
+
+    @classmethod
+    def _parse(cls, elem):
+        res = super()._parse(elem)
+        res.update({
+            'shuffle_options': bool(list_get(cls._get_entry(elem), cls._Index.SHUFFLE_OPTIONS, False)),
+        })
+        return res
+
+    def __init__(self, *, shuffle_options, **kwargs):
+        super().__init__(**kwargs)
+        self.shuffle_options = shuffle_options
+
+
     @property
     def options(self):
         return self._options[0]
@@ -427,7 +476,6 @@ class OtherChoiceInput(ChoiceInput1D):
 
     def payload(self) -> Dict[str, List[str]]:
         """See base class."""
-
         payload = super().payload()
         # '' is not converted to None (see _validate_entry), so "... not None" isn't applicable
         if not self._other_value:
@@ -441,7 +489,6 @@ class OtherChoiceInput(ChoiceInput1D):
 
     def draft(self) -> List[Tuple]:
         """See base class."""
-
         result = super().draft()
         if not self._other_value:
             return result
@@ -454,6 +501,9 @@ class OtherChoiceInput(ChoiceInput1D):
         result[0] = (*data, answers, 1)
         return result
 
+    def _is_empty(self):
+        return super()._is_empty() and self._other_value is None
+
     def _set_choices(self, choices: List[Union[List[ChoiceValue], EmptyValue]]):
         self._other_value = None
         super()._set_choices(choices)
@@ -502,18 +552,82 @@ class OtherChoiceInput(ChoiceInput1D):
         return hints
 
 
-class Grid(ChoiceInput):
+class ValidatedInput(InputElement, ABC):
+    """An input element which may have a validator.
+
+    Attributes:
+        validator: The validator, if it exists.
+    """
+
+    @classmethod
+    def _parse(cls, elem):
+        res = super()._parse(elem)
+        res.update({
+            'validator': cls._parse_validator(elem)
+        })
+        return res
+
+    def __init__(self, *, validator, **kwargs):
+        super().__init__(**kwargs)
+        self.validator: Optional[Validator] = validator
+
+    def is_misconfigured(self):
+        """Checks if the element cannot be filled with a valid value."""
+        if self.validator is None:
+            return False
+        if self.validator.has_unknown_type() or self.validator.bad_args:
+            return False
+        if not self.required:  # always can set empty value
+            return False
+        return self._is_misconfigured()
+
+    def _validate(self):
+        """See base validate method.
+
+        If this element is required and MisconfiguredElement is raised,
+        you should skip the page with this element, if possible.
+        Otherwise, the form is not submittable.
+
+        Raises:
+            gforms.errors.MisconfiguredElement:
+                This element will not accept any non-empty value.
+        """
+        if self.is_misconfigured():
+            raise MisconfiguredElement(self)
+        super()._validate()
+        if self.validator is not None and not self._is_empty():  # Non-empty value
+            self.validator.validate(self)
+
+    def _hints(self, indent=0, modify=False):
+        res = []
+        if self.validator is not None:
+            res.append(f'! {self.validator.to_str()} !')
+        return res + super()._hints(indent, modify)
+
+    @classmethod
+    @abstractmethod
+    def _parse_validator(cls, elem) -> Optional[Validator]:
+        raise NotImplementedError()
+
+    @abstractmethod
+    def _is_misconfigured(self):
+        # NOTE method is called only when self.validator is not None and self.required is True
+        raise NotImplementedError()
+
+
+class Grid(ValidatedInput, ChoiceInput):
     """A Grid input element.
 
     Attributes:
         rows: The grid row names.
         cols: The grid columns (an alias for options).
-        validator: The grid validator, if it exists.
+        shuffle_rows: Shuffle this grid's rows.
     """
 
     _cell_symbols = ('?', '+')
 
-    class _Index(InputElement._Index):
+    class _Index(ChoiceInput._Index):
+        SHUFFLE_ROWS = 7
         VALIDATOR = 8
         ROW_NAME = 3
         MULTICHOICE = 11
@@ -526,32 +640,34 @@ class Grid(ChoiceInput):
     def _parse(cls, elem):
         res = super()._parse(elem)
         res.update({
-            'rows': [entry[cls._Index.ROW_NAME][0] for entry in cls._get_entries(elem)]
+            'rows': [entry[cls._Index.ROW_NAME][0] for entry in cls._get_entries(elem)],
+            'shuffle_rows': bool(list_get(elem, cls._Index.SHUFFLE_ROWS, False)),
         })
-        if len(elem) > cls._Index.VALIDATOR:
-            res['validator'] = GridValidator.parse(elem[cls._Index.VALIDATOR])
         return res
 
-    def __init__(self, *, rows, validator=None, **kwargs):
+    @classmethod
+    def _parse_validator(cls, elem) -> Optional[GridValidator]:
+        if len(elem) > cls._Index.VALIDATOR and elem[cls._Index.VALIDATOR]:
+            return GridValidator.parse(elem[cls._Index.VALIDATOR][0])
+        return None
+
+    def __init__(self, *, rows, shuffle_rows, **kwargs):
         super().__init__(**kwargs)
         self.rows = rows
         self.cols = self.options  # alias
-        self.validator: Optional[GridValidator] = validator
+        self.shuffle_rows = shuffle_rows
 
     @property
     def options(self):
         return self._options[0]
 
-    def is_misconfigured(self):
-        """Checks if the grid can never be filled with valid values."""
-
-        return self.required and len(self.cols) < len(self.rows)
+    def _is_misconfigured(self):
+        return self.validator.subtype is GridTypes.EXCLUSIVE_COLUMNS \
+            and len(self.cols) < len(self.rows)
 
     def _hints(self, indent=0, modify=False):
         # NOTE row/column names may need wrapping
-        res = []
-        if self.validator is not None:
-            res.append(self.validator.to_str())
+        res = super()._hints(indent, modify)
         max_length = max(len(row) for row in self.rows)
         header = ' ' * (max_length + 1) + '|'.join([opt.value for opt in self.options])
         row_fmt = f'{{:>{max_length}}} ' + \
@@ -599,15 +715,8 @@ class Grid(ChoiceInput):
         except RequiredElement as e:
             raise RequiredRow(self, index=e.index)
 
-    def validate(self):
-        """See base class."""
 
-        super().validate()
-        if self.validator is not None:
-            self.validator.validate(self, self._values)
-
-
-class TimeElement(SingleInput):
+class TimeInput(SingleInput):
     """An element which represents a time or a duration."""
 
     class _Index(SingleInput._Index):
@@ -616,8 +725,8 @@ class TimeElement(SingleInput):
 
     @staticmethod
     def parse_duration_flag(elem):
-        flags = TimeElement._get_entry(elem)[TimeElement._Index.FLAGS]
-        return bool(flags[TimeElement._Index.DURATION])
+        flags = TimeInput._get_entry(elem)[TimeInput._Index.FLAGS]
+        return bool(flags[TimeInput._Index.DURATION])
 
     def __init__(self, **kwargs):
         super().__init__(**kwargs)
@@ -625,15 +734,8 @@ class TimeElement(SingleInput):
         self._minute = None
         self._second = None
 
-    def validate(self):
-        """See base class."""
-
-        if self.required and self._is_empty():
-            raise RequiredElement(self)
-
     def payload(self) -> Dict[str, List[str]]:
         """See base class."""
-
         payload = {}
         if self._hour is not None:
             payload[self._part_id('hour')] = self._hour
@@ -645,7 +747,6 @@ class TimeElement(SingleInput):
 
     def draft(self) -> List[Tuple]:
         """See base class."""
-
         hms = []
         for val in [self._hour, self._minute, self._second]:
             if val is not None:
@@ -654,6 +755,10 @@ class TimeElement(SingleInput):
         # The server sets the same draft value, so it's ok
         return [(None, self._entry_id, [':'.join(hms)], 0)]
 
+    def _validate(self):
+        if self.required and self._is_empty():
+            raise RequiredElement(self)
+
     def _set_time(self, hour, minute, second):
         self._hour = hour
         self._minute = minute
@@ -663,7 +768,7 @@ class TimeElement(SingleInput):
         return all(val is None for val in [self._hour, self._minute, self._second])
 
 
-class DateElement(SingleInput):
+class DateInput(SingleInput):
     """An element which represents a date.
 
     Attributes:
@@ -677,8 +782,8 @@ class DateElement(SingleInput):
 
     @staticmethod
     def parse_time_flag(elem):
-        flags = DateElement._get_entry(elem)[DateElement._Index.FLAGS]
-        return bool(flags[DateElement._Index.TIME])
+        flags = DateInput._get_entry(elem)[DateInput._Index.FLAGS]
+        return bool(flags[DateInput._Index.TIME])
 
     @classmethod
     def _parse(cls, elem):
@@ -694,15 +799,8 @@ class DateElement(SingleInput):
         self.has_year = has_year
         self._date = None
 
-    def validate(self):
-        """See base class."""
-
-        if self.required and self._date is None:
-            raise RequiredElement(self)
-
     def payload(self) -> Dict[str, List[str]]:
         """See base class."""
-
         if self._date is None:
             return {}
         payload = {
@@ -715,12 +813,15 @@ class DateElement(SingleInput):
 
     def draft(self) -> List[Tuple]:
         """See base class."""
-
         if self._date is None:
             return []
         fmt = '%Y-%m-%d' if self.has_year else '%m-%d'
         return [(None, self._entry_id, [self._date.strftime(fmt)], 0)]
 
+    def _validate(self):
+        if self.required and self._date is None:
+            raise RequiredElement(self)
+
     def _answer(self) -> List[str]:
         if self._date is None:
             return ['EMPTY']
@@ -729,6 +830,9 @@ class DateElement(SingleInput):
             fmt = '%Y/' + fmt
         return [f'"{self._date.strftime(fmt)}"']
 
+    def _is_empty(self):
+        return self._date is None
+
 
 class MediaElement(Element):
     """A base class for media elements."""
@@ -736,27 +840,18 @@ class MediaElement(Element):
     pass
 
 
-class TextInput(SingleInput):
-    """An element which accepts text input.
-
-    Attributes:
-        validator: The validator for this element, if it exists.
-    """
+class TextInput(ValidatedInput, SingleInput):
+    """An element which accepts text input."""
 
     class _Index(SingleInput._Index):
         VALIDATOR = 4
 
     @classmethod
-    def _parse(cls, elem):
-        res = super()._parse(elem)
+    def _parse_validator(cls, elem) -> Optional[TextValidator]:
         value = cls._get_entry(elem)
-        if len(value) > cls._Index.VALIDATOR:
-            res['validator'] = TextValidator.parse(value[cls._Index.VALIDATOR][0])
-        return res
-
-    def __init__(self, *, validator=None, **kwargs):
-        super().__init__(**kwargs)
-        self.validator: Optional[TextValidator] = validator
+        if len(value) > cls._Index.VALIDATOR and value[cls._Index.VALIDATOR]:
+            return TextValidator.parse(value[cls._Index.VALIDATOR][0])
+        return None
 
     def set_value(self, value: Union[TextValue, EmptyValue]):
         """Sets the value for this element.
@@ -770,7 +865,6 @@ class TextInput(SingleInput):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         if isinstance(value, str):
             if not value:
                 value = Value.EMPTY
@@ -780,15 +874,9 @@ class TextInput(SingleInput):
             return self._set_value(value)
         raise ElementTypeError(self, value)
 
-    def _validate_entry(self, index):  # index == 0
-        super()._validate_entry(index)
-        if self.validator is not None and self._values[index]:
-            self.validator.validate(self, self._values[index][0])
-
-    def _hints(self, indent=0, modify=False):
-        if self.validator is not None:
-            return [self.validator.to_str()]
-        return []
+    def _is_misconfigured(self):
+        return self.validator.subtype is NumberTypes.RANGE \
+            and self.validator.args[0] > self.validator.args[1]
 
 
 class ActionChoiceInput(ChoiceInput1D):
@@ -814,12 +902,16 @@ class ActionChoiceInput(ChoiceInput1D):
             gforms.errors.ElementTypeError:
                 The argument's type is not accepted by this element.
         """
-
         return self._set_choices([self._to_choice_list(value)])
 
     def _set_choices(self, choices: List[Union[List[ChoiceValue], EmptyValue]]):
         self.next_page = None
         super()._set_choices(choices)
+        # Don't reset no_loops if the element's options have no actions or the actions are ignored
+        if (self._form is not None and
+                isinstance(self.options[0], ActionOption) and
+                cast(ActionOption, self.options[0]).next_page is not None):
+            self._form._no_loops = False
 
     def _add_choice(self, choices, choice: Option):
         super()._add_choice(choices, choice)
