@@ -4,7 +4,7 @@ from abc import ABC, abstractmethod
 class FormsError(Exception, ABC):
     """Base error class"""
 
-    def __init__(self, *args, details=None, **kwargs):
+    def __init__(self, *, details=None):
         super().__init__()
         self.details = details
 
@@ -19,30 +19,62 @@ class FormsError(Exception, ABC):
         return message
 
 
+class InvalidURL(FormsError):
+    def __init__(self, url, *args, **kwargs):
+        super().__init__(*args, **kwargs)
+        self.url = url
+
+    def _message(self):
+        return f'Not a valid form URL: "{self.url}".'
+
+
+class NoSuchForm(InvalidURL):
+    def _message(self):
+        return f'A form with URL "{self.url}" does not exist.'
+
+
 class FormError(FormsError):
     def __init__(self, form, *args, **kwargs):
         super().__init__(*args, **kwargs)
         self.form = form
 
+    @property
+    def form_descr(self):
+        if self.form.title is not None:
+            return f'"{self.form.title}"'
+        elif self.form.url is not None:
+            return 'with URL ' + self.form.url
+        return ''
+
 
 class ParseError(FormError):
     def _message(self):
-        return f'Cannot parse form with URL {self.form.url}'
+        return 'Cannot parse form ' + self.form_descr
 
 
 class ClosedForm(ParseError):
     def _message(self):
-        return f'Form "{self.form.title}" is closed'
+        return f'Form {self.form_descr} is closed'
+
+
+class SigninRequired(FormError, NotImplementedError):
+    def _message(self):
+        return f'Form {self.form_descr} requires sign in'
+
+
+class EditingDisabled(ParseError):
+    def _message(self):
+        return 'Cannot edit a response for form ' + self.form_descr
 
 
 class FormNotLoaded(FormError):
     def _message(self):
-        return f'Form with URL "{self.form.url}" was not loaded'
+        return f'Form {self.form_descr} was not loaded'
 
 
-class FormNotFilled(FormError):
+class FormNotValidated(FormError):
     def _message(self):
-        return f'Form "{self.form.title}" was not filled'
+        return f'The form needs to be validated.'
 
 
 class ElementError(FormsError):
@@ -52,7 +84,7 @@ class ElementError(FormsError):
         self.index = index
 
 
-class InfiniteLoop(FormsError, ValueError):
+class InfiniteLoop(FormError, ValueError):
     def __init__(self, form):
         super().__init__(form)
 
@@ -66,16 +98,22 @@ class RowError(ElementError):
         return self.elem.rows[self.index]
 
 
-class ValidationError(FormsError, ValueError):
+class ValidationError(ElementError, ValueError):
     pass
 
 
-class InvalidValue(ElementError, ValidationError):
+class InvalidValue(ValidationError):
     def __init__(self, elem, value, *args, **kwargs):
         super().__init__(elem, *args, **kwargs)
         self.value = value
 
 
+class MisconfiguredElement(ValidationError):
+    def _message(self):
+        return f'The validator requirements for element "{self.elem.name}" are:' \
+               f' "{self.elem.validator.to_str()}". A valid value for this element does not exist.'
+
+
 # The following errors may be raised from InputElement.set_value
 
 class ElementTypeError(ElementError, TypeError):
@@ -126,7 +164,7 @@ class InvalidRowChoice(InvalidChoice, RowError):
 
 # The following errors may be raised from InputElement.validate
 
-class RequiredElement(ElementError, ValidationError):
+class RequiredElement(ValidationError):
     def _message(self):
         return f'An entry in required element "{self.elem.name}" is empty'
 
@@ -152,6 +190,16 @@ class InvalidDuration(InvalidValue):
                ' value should be positive and less than 73 hours'
 
 
+class UnknownElement(UserWarning):
+    def __init__(self, el_data, el_type):
+        super().__init__()
+        self.el_data = el_data
+        self.el_type = el_type
+
+    def __str__(self):
+        return f'Found an element with unknown type ({self.el_type}). Raw element data: {self.el_data}'
+
+
 class ValidatorWarning(UserWarning):
     pass
 
@@ -168,21 +216,22 @@ class UnknownValidator(ValidatorWarning):
 
 
 class InvalidArguments(ValidatorWarning):
-    def __init__(self, val_type, subtype, args):
+    def __init__(self, cls, val_type, subtype, args):
         super().__init__()
+        self.cls = cls
         self.val_type = val_type
         self.subtype = subtype
         self.args = args
 
     def __str__(self):
-        return 'Cannot convert arguments for a validator.' \
-              f' Type: {self.val_type} ({self.subtype}), Arguments: {repr(self.args)}.' \
+        return f'Cannot convert arguments for {self.cls.__name__}.' \
+               f' Type: {self.val_type} ({self.subtype}), Arguments: {repr(self.args)}.' \
                ' Validator is disabled'
 
 
-class SameColumn(InvalidValue):
-    def __init__(self, elem, value=None, column=None, *args, **kwargs):
-        super().__init__(elem, value, *args, **kwargs)
+class SameColumn(ValidationError):
+    def __init__(self, elem, column=None, *args, **kwargs):
+        super().__init__(elem, *args, **kwargs)
         self.column = column
 
     def _message(self):
@@ -191,7 +240,12 @@ class SameColumn(InvalidValue):
                ' The choices must be unique'
 
 
-class MisconfiguredGrid(SameColumn):
+class InvalidChoiceCount(ValidationError):
+    def __init__(self, elem, cnt, *args, **kwargs):
+        super().__init__(elem, *args, **kwargs)
+        self.cnt = cnt
+
     def _message(self):
-        return f'Impossible to fill grid "{self.elem}": all rows are required and the choices' \
-               ' must be unique, but there are less columns than rows'
+        val = self.elem.validator
+        return f'{self.cnt} option(s) were chosen in "{self.elem.name}",' \
+               f' but {val.subtype.descr(val.args)} are required'
