@@ -1,3 +1,6 @@
+from .util import list_get
+
+
 class Option:
     """A choice option for a ChoiceInput.
 
@@ -18,22 +21,14 @@ class Option:
 
         This method should not be called directly,, use options.parse instead.
         """
-
         return cls(**cls._parse(option))
 
     @classmethod
     def _parse(cls, option):
-        res = {
-            'value': option[cls._Index.VALUE],
-            'other': False,
+        return {
+            'value': option[cls._Index.VALUE] or '',
+            'other': bool(list_get(option, cls._Index.OTHER, False)),
         }
-        # len(option) == 1 for Scale options or if the element has only one option with no actions
-        if len(option) > cls._Index.OTHER:
-            res.update({
-                'other': bool(option[cls._Index.OTHER]),
-                'value': res['value'] or ''
-            })
-        return res
 
     def __init__(self, *, value, other):
         self.value = value
@@ -44,7 +39,6 @@ class Option:
 
         For args description, see Form.to_str.
         """
-
         if self.other:
             if with_value is not None:
                 return f'Other: "{with_value}"'
@@ -74,7 +68,6 @@ class ActionOption(Option):
 
     def to_str(self, indent=0, with_value=None):
         """See base class."""
-
         from .elements import Page
         s = super().to_str(indent, with_value)
         if self.next_page is Page.SUBMIT:
@@ -95,7 +88,6 @@ class ActionOption(Option):
 
 def parse(option):
     """Creates an Option of the right type from its JSON representation."""
-
     if len(option) > Option._Index.ACTION and option[Option._Index.ACTION] is not None:
         return ActionOption.parse(option)
     return Option.parse(option)
